@@ -5,10 +5,6 @@
 #
 import io
 import logging
-#from pypykatz.commons.common import *
-#from pypykatz.crypto.des import *
-#from pypykatz.crypto.aes import AESModeOfOperationCBC
-#from pypykatz.lsadecryptor.lsa_templates import *
 from pypykatz.crypto.RC4 import RC4
 from pypykatz.lsadecryptor.package_commons import PackageDecryptor
 from pypykatz.commons.win_datatypes import LONG
@@ -92,7 +88,7 @@ class LsaDecryptor_NT5(PackageDecryptor):
 		self.log('Looking for main struct signature in memory...')
 		fl = self.reader.find_in_module('lsasrv.dll', self.decryptor_template.signature)
 		if len(fl) == 0:
-			logging.warning('signature not found! %s' % self.decryptor_template.signature.hex())
+			logging.debug('signature not found! %s' % self.decryptor_template.signature.hex())
 			raise Exception('LSA signature not found!')
 			
 		self.log('Found candidates on the following positions: %s' % ' '.join(hex(x) for x in fl))
