@@ -4,10 +4,12 @@
 #  Tamas Jos (@skelsec)
 #
 import io
-from minidump.win_datatypes import *
-from pypykatz.commons.common import *
-from pypykatz.commons.win_datatypes import *
-from pypykatz.lsadecryptor.package_commons import *
+from minidump.win_datatypes import ULONG64, FILETIME, PCWSTR, SIZE_T, BOOL  #WCHAR, ANYSIZE_ARRAY
+from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
+from pypykatz.commons.win_datatypes import POINTER, PVOID, ULONG, LIST_ENTRY, \
+	DWORD, LSA_UNICODE_STRING, PKERB_EXTERNAL_NAME, KIWI_GENERIC_PRIMARY_CREDENTIAL, \
+	LUID, PLSAISO_DATA_BLOB
+from pypykatz.lsadecryptor.package_commons import PackageTemplate
 
 class KerberosTemplate(PackageTemplate):
 	def __init__(self, sysinfo):
@@ -32,7 +34,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_51
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
 				template.hash_password_struct = KERB_HASHPASSWORD_5
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_5
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
 				
 				
 			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
@@ -42,7 +44,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_52
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
 				template.hash_password_struct = KERB_HASHPASSWORD_5
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_5
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
 				
 			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
 				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
@@ -51,7 +53,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_60
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_60
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
 				
 			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
 				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
@@ -60,7 +62,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_60
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
 			
 			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
 				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
@@ -69,7 +71,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_62
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_62
 				
 			elif WindowsBuild.WIN_10_1507.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1511.value:
 				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
@@ -78,7 +80,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_10
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
 				
 			elif WindowsBuild.WIN_10_1511.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1607.value:
 				template.signature = b'\x48\x8b\x18\x48\x8d\x0d'
@@ -87,7 +89,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_10
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
 				
 
 			elif sysinfo.buildnumber >= WindowsBuild.WIN_10_1607.value:
@@ -97,10 +99,10 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_10
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
 			
 			else:
-				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (architecture, sysinfo.buildnumber))
+				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
 			
 		
 		elif sysinfo.architecture == KatzSystemArchitecture.X86:
@@ -111,7 +113,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_51
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
 				template.hash_password_struct = KERB_HASHPASSWORD_5
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_5
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
 				
 				
 			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
@@ -121,7 +123,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_52
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_5
 				template.hash_password_struct = KERB_HASHPASSWORD_5
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_5
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_5
 				
 			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
 				template.signature = b'\x53\x8b\x18\x50\x56'
@@ -130,7 +132,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_60
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_60
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
 				
 			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
 				template.signature = b'\x53\x8b\x18\x50\x56'
@@ -139,7 +141,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_60
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_60
 				
 			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsBuild.WIN_BLUE.value:
 				template.signature = b'\x57\x8b\x38\x50\x68'
@@ -148,7 +150,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_62
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_62
 				
 			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
 				template.signature = b'\x56\x8b\x30\x50\x57'
@@ -157,7 +159,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_62
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_62
 			
 			####DOUBLE CHECK THE STRUCTURES BELOW THIS LINE!!!!
 			#### kerbHelper[N] -> KerberosReferences... {-15,7}}, here N= 7
@@ -169,7 +171,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_6
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_10
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
 				
 				
 			elif WindowsBuild.WIN_10_1511.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1903.value:
@@ -179,7 +181,7 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_10
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
 				
 				
 			elif WindowsBuild.WIN_10_1903.value <= sysinfo.buildnumber:
@@ -189,11 +191,11 @@ class KerberosTemplate(PackageTemplate):
 				template.kerberos_ticket_struct = KIWI_KERBEROS_INTERNAL_TICKET_10_1607
 				template.keys_list_struct = KIWI_KERBEROS_KEYS_LIST_6
 				template.hash_password_struct = KERB_HASHPASSWORD_6_1607
-				template.csp_info_struct = PKIWI_KERBEROS_CSP_INFOS_10
+				template.csp_info_struct = KIWI_KERBEROS_CSP_INFOS_10
 		
 		
 		else:
-			raise Exception('Unknown architecture! %s' % architecture)
+			raise Exception('Unknown architecture! %s' % sysinfo.architecture)
 
 			
 		return template
@@ -204,14 +206,45 @@ class PKERB_SMARTCARD_CSP_INFO_5(POINTER):
 		
 
 class KERB_SMARTCARD_CSP_INFO_5:
-	def __init__(self, reader):
-		self.dwCspInfoLen = DWORD(reader).value
+	def __init__(self, reader, size):
+		pos = reader.tell()
+		#self.dwCspInfoLen = DWORD(reader).value
 		self.ContextInformation = PVOID(reader).value
 		self.nCardNameOffset = ULONG(reader).value
 		self.nReaderNameOffset = ULONG(reader).value
 		self.nContainerNameOffset = ULONG(reader).value
 		self.nCSPNameOffset = ULONG(reader).value
-		self.bBuffer = WCHAR(reader).value
+		diff = reader.tell() - pos
+		data = reader.read(size - diff + 4)
+		self.bBuffer = io.BytesIO(data)
+
+	def read_wcharnull(self, buffer, tpos):
+		pos = buffer.tell()
+		buffer.seek(tpos, 0)
+		data = b''
+		i=0
+		nc = 0
+		while i < 255:
+			if nc == 3:
+				break
+			c = buffer.read(1)
+			if c == b'\x00':
+				nc += 1
+			else:
+				nc = 0
+			data += c
+			i += 1
+		buffer.seek(pos, 0)
+		return data.decode('utf-16-le').replace('\x00', '')
+
+	def get_infos(self):
+		t = {}
+		t['CardName'] = self.read_wcharnull(self.bBuffer, self.nCardNameOffset)
+		t['ReaderName'] = self.read_wcharnull(self.bBuffer, self.nReaderNameOffset)
+		t['ContainerName'] = self.read_wcharnull(self.bBuffer, self.nContainerNameOffset)
+		t['CSPName'] = self.read_wcharnull(self.bBuffer, self.nCSPNameOffset)
+
+		return t
 
 class PKERB_SMARTCARD_CSP_INFO(POINTER):
 	def __init__(self, reader):
@@ -219,18 +252,49 @@ class PKERB_SMARTCARD_CSP_INFO(POINTER):
 		
 		
 class KERB_SMARTCARD_CSP_INFO:
-	def __init__(self, reader):
-		self.dwCspInfoLen = DWORD(reader).value
+	def __init__(self, reader, size):
+		pos = reader.tell()
+		#self.dwCspInfoLen = DWORD(reader).value
 		self.MessageType = DWORD(reader).value
 		self.ContextInformation = PVOID(reader).value #U
 		self.SpaceHolderForWow64 = ULONG64(reader).value #U
 		self.flags = DWORD(reader).value
 		self.KeySpec = DWORD(reader).value
-		self.nCardNameOffset = ULONG(reader).value
-		self.nReaderNameOffset = ULONG(reader).value
-		self.nContainerNameOffset = ULONG(reader).value
-		self.nCSPNameOffset = ULONG(reader).value
-		self.bBuffer[ANYSIZE_ARRAY] = WCHAR(reader).value
+		self.nCardNameOffset = ULONG(reader).value * 2
+		self.nReaderNameOffset = ULONG(reader).value * 2
+		self.nContainerNameOffset = ULONG(reader).value * 2
+		self.nCSPNameOffset = ULONG(reader).value * 2
+		diff = reader.tell() - pos
+		data = reader.read(size - diff + 4)
+		self.bBuffer = io.BytesIO(data)
+
+	def read_wcharnull(self, buffer, tpos):
+		pos = buffer.tell()
+		buffer.seek(tpos, 0)
+		data = b''
+		i=0
+		nc = 0
+		while i < 255:
+			if nc == 3:
+				break
+			c = buffer.read(1)
+			if c == b'\x00':
+				nc += 1
+			else:
+				nc = 0
+			data += c
+			i += 1
+		buffer.seek(pos, 0)
+		return data.decode('utf-16-le').replace('\x00', '')
+
+	def get_infos(self):
+		t = {}
+		t['CardName'] = self.read_wcharnull(self.bBuffer, self.nCardNameOffset)
+		t['ReaderName'] = self.read_wcharnull(self.bBuffer, self.nReaderNameOffset)
+		t['ContainerName'] = self.read_wcharnull(self.bBuffer, self.nContainerNameOffset)
+		t['CSPName'] = self.read_wcharnull(self.bBuffer, self.nCSPNameOffset)
+
+		return t
 		
 class PKIWI_KERBEROS_CSP_INFOS_5(POINTER):
 	def __init__(self, reader):
@@ -239,13 +303,13 @@ class PKIWI_KERBEROS_CSP_INFOS_5(POINTER):
 class KIWI_KERBEROS_CSP_INFOS_5:
 	def __init__(self, reader):	
 		self.PinCode = LSA_UNICODE_STRING(reader)
-		self.unk0 = PVOID(reader).value
-		self.unk1 = PVOID(reader).value
-		self.CertificateInfos = PVOID(reader).value
-		self.unkData = PVOID(reader).value                #	// 0 = CspData
+		self.unk0 = PVOID(reader)
+		self.unk1 = PVOID(reader)
+		self.CertificateInfos = PVOID(reader)
+		self.unkData = PVOID(reader)                      #	// 0 = CspData
 		self.Flags = DWORD(reader).value                  #	// 1 = CspData (not 0x21)(reader).value
 		self.CspDataLength = DWORD(reader).value
-		self.CspData = KERB_SMARTCARD_CSP_INFO_5(reader).value
+		self.CspData = KERB_SMARTCARD_CSP_INFO_5(reader, size = self.CspDataLength)
 	 
 class PKIWI_KERBEROS_CSP_INFOS_60(POINTER):
 	def __init__(self, reader):
@@ -262,7 +326,7 @@ class KIWI_KERBEROS_CSP_INFOS_60:
 		self.Flags = DWORD(reader).value	            #// 0 = CspData(reader).value
 		self.unkFlags = DWORD(reader).value      	#// 0x141(reader).value
 		self.CspDataLength = DWORD(reader).value
-		self.CspData = KERB_SMARTCARD_CSP_INFO(reader).value
+		self.CspData = KERB_SMARTCARD_CSP_INFO(reader, size = self.CspDataLength)
 
 class PKIWI_KERBEROS_CSP_INFOS_62(POINTER):
 	def __init__(self, reader):
@@ -280,7 +344,7 @@ class KIWI_KERBEROS_CSP_INFOS_62:
 		self.Flags = DWORD(reader).value	             #// 0 = CspData(reader).value
 		self.unkFlags = DWORD(reader).value	            #// 0x141 (not 0x61)
 		self.CspDataLength = DWORD(reader).value
-		self.CspData = KERB_SMARTCARD_CSP_INFO(reader).value
+		self.CspData = KERB_SMARTCARD_CSP_INFO(reader, size = self.CspDataLength)
 		
 class PKIWI_KERBEROS_CSP_INFOS_10(POINTER):
 	def __init__(self, reader):
@@ -298,7 +362,7 @@ class KIWI_KERBEROS_CSP_INFOS_10:
 		self.unkFlags = DWORD(reader).value	        #// 0x141 (not 0x61)(reader).value
 		self.unk3 = PVOID(reader).value
 		self.CspDataLength = DWORD(reader).value
-		self.CspData = KERB_SMARTCARD_CSP_INFO(reader).value
+		self.CspData = KERB_SMARTCARD_CSP_INFO(reader, size = self.CspDataLength)
 
 class PKIWI_KERBEROS_LOGON_SESSION_51(POINTER):
 	def __init__(self, reader):
@@ -339,7 +403,7 @@ class KIWI_KERBEROS_LOGON_SESSION_51:
 		self.Tickets_1 = LIST_ENTRY(reader)
 		self.Tickets_2 = LIST_ENTRY(reader)
 		self.Tickets_3 = LIST_ENTRY(reader)
-		self.SmartcardInfos = PVOID(reader).value
+		self.SmartcardInfos = PVOID(reader)
 
 		
 class PKIWI_KERBEROS_LOGON_SESSION(POINTER):
@@ -385,7 +449,7 @@ class KIWI_KERBEROS_LOGON_SESSION:
 		self.unk25 = FILETIME(reader).value
 		self.Tickets_3 = LIST_ENTRY(reader)
 		self.unk26 = FILETIME(reader).value
-		self.SmartcardInfos = PVOID(reader).value
+		self.SmartcardInfos = PVOID(reader)
 
 class PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL(POINTER):
 	def __init__(self, reader):
@@ -456,7 +520,7 @@ class KIWI_KERBEROS_LOGON_SESSION_10_X86:
 		self.unk28 = FILETIME(reader).value
 		self.Tickets_3 = LIST_ENTRY(reader)
 		self.unk29 = FILETIME(reader).value
-		self.SmartcardInfos = PVOID(reader).value
+		self.SmartcardInfos = PVOID(reader)
 		
 class KIWI_KERBEROS_LOGON_SESSION_10:
 	def __init__(self, reader):	
@@ -500,7 +564,7 @@ class KIWI_KERBEROS_LOGON_SESSION_10:
 		self.unk28 = FILETIME(reader).value
 		self.Tickets_3 = LIST_ENTRY(reader)
 		self.unk29 = FILETIME(reader).value
-		self.SmartcardInfos = PVOID(reader).value
+		self.SmartcardInfos = PVOID(reader)
 
 class PKIWI_KERBEROS_10_PRIMARY_CREDENTIAL_1607_ISO(POINTER):
 	def __init__(self, reader):
@@ -579,7 +643,7 @@ class KIWI_KERBEROS_LOGON_SESSION_10_1607:
 		self.unk28 = FILETIME(reader).value
 		self.Tickets_3 = LIST_ENTRY(reader)
 		self.unk29 = FILETIME(reader).value
-		self.SmartcardInfos = PVOID(reader).value
+		self.SmartcardInfos = PVOID(reader)
 		
 		
 class KIWI_KERBEROS_LOGON_SESSION_10_1607_X86:
@@ -631,7 +695,7 @@ class KIWI_KERBEROS_LOGON_SESSION_10_1607_X86:
 		self.unk28 = FILETIME(reader).value
 		self.Tickets_3 = LIST_ENTRY(reader)
 		self.unk29 = FILETIME(reader).value
-		self.SmartcardInfos = PVOID(reader).value
+		self.SmartcardInfos = PVOID(reader)
 
 class PKIWI_KERBEROS_INTERNAL_TICKET_51(POINTER):
 	def __init__(self, reader):
@@ -953,7 +1017,7 @@ class KIWI_KERBEROS_KEYS_LIST_5:
 
 	def read(self, reader, keyentries_type):
 		reader.move(self.KeyEntries_start)
-		for i in range(self.cbItem):
+		for _ in range(self.cbItem):
 			self.KeyEntries.append(keyentries_type(reader))
 
 class PKIWI_KERBEROS_KEYS_LIST_6(POINTER):
@@ -974,7 +1038,7 @@ class KIWI_KERBEROS_KEYS_LIST_6:
 		
 	def read(self, reader, keyentries_type):
 		reader.move(self.KeyEntries_start)
-		for i in range(self.cbItem):
+		for _ in range(self.cbItem):
 			self.KeyEntries.append(keyentries_type(reader))
 			#//KERB_HASHPASSWORD_6 KeysEntries[ANYSIZE_ARRAY] = (reader).value
 
